@@ -1,19 +1,23 @@
 from logging import getLogger
 
 from aiohttp.web import json_response
-from django.contrib.postgres.fields import ArrayField, JSONField
+from django.contrib.postgres.fields import ArrayField
+from django.core.exceptions import ObjectDoesNotExist
 from django.db import models
+from django.conf import settings
+from yarl import URL
 
 from pulpcore.plugin.models import (
     Content,
     Publication,
-    PublicationDistribution,
+    Distribution,
     Remote,
     Repository
 )
 
 from pathlib import PurePath
 from .utils import python_content_to_json, PYPI_LAST_SERIAL, PYPI_SERIAL_CONSTANT
+from pulpcore.plugin.repo_version_utils import remove_duplicates, validate_repo_version
 
 log = getLogger(__name__)
 
@@ -35,13 +39,15 @@ PLATFORMS = (("windows", "windows"),
              ("linux", "linux"))
 
 
-class PythonDistribution(PublicationDistribution):
+class PythonDistribution(Distribution):
     """
     Distribution for 'Python' Content.
     """
 
     TYPE = 'python'
 
+    allow_uploads = models.BooleanField(default=True)
+
     def content_handler(self, path):
         """
         Handler to serve extra, non-Artifact content for this Distribution
@@ -60,6 +66,38 @@ class PythonDistribution(PublicationDistribution):
             name = path.parts[1]
         elif path.match("pypi/*/json"):
             name = path.parts[1]
+        elif len(path.parts) and path.parts[0] == "simple":
+            # Temporary fix for PublishedMetadata not being properly served from remote storage
+            # https://github.com/pulp/pulp_python/issues/413
+            if settings.DEFAULT_FILE_STORAGE != "pulpcore.app.models.storage.FileSystem":
+                try:
+                    publication = self.publication or Publication.objects.filter(
+                        repository_version=self.repository.latest_version()).latest("pulp_created")
+                except ObjectDoesNotExist:
+                    return None
+                rel_path = f"{path}/index.html"
+                try:
+                    ca = publication.published_artifact.select_related(
+                        "content_artifact",
+                        "content_artifact__artifact",
+                    ).get(relative_path=rel_path).content_artifact
+                except ObjectDoesNotExist:
+                    return None
+                file = ca.artifact.file
+                content_disposition = f"attachment;filename={ca.relative_path}"
+                parameters = {
+                    "ResponseContentDisposition": content_disposition,
+                    "ResponseContentType": "text/html"
+                }
+                url = URL(file.storage.url(file.name, parameters=parameters), encoded=True)
+                # Trick the content app to stream the metadata from the remote storage
+                remote = PythonRemote(name="Redirect", url=str(url), policy="streamed")
+                remote.get_remote_artifact_url = lambda x: str(url)
+                setattr(self, "publication", None)
+                setattr(self, "repository", None)
+                setattr(self, "remote", remote)
+                return None
+
         if name:
             package_content = PythonPackageContent.objects.filter(
                 pk__in=self.publication.repository_version.content,
@@ -87,12 +125,16 @@ class PythonPackageContent(Content):
     https://www.python.org/dev/peps/pep-0345/
     """
 
+    PROTECTED_FROM_RECLAIM = False
+
     TYPE = 'python'
+    repo_key_fields = ("filename",)
     # Required metadata
-    filename = models.TextField(unique=True, db_index=True)
+    filename = models.TextField(db_index=True)
     packagetype = models.TextField(choices=PACKAGE_TYPES)
     name = models.TextField()
     version = models.TextField()
+    sha256 = models.CharField(unique=True, db_index=True, max_length=64)
     # Optional metadata
     python_version = models.TextField()
     metadata_version = models.TextField()
@@ -110,11 +152,13 @@ class PythonPackageContent(Content):
     project_url = models.TextField()
     platform = models.TextField()
     supported_platform = models.TextField()
-    requires_dist = JSONField(default=list)
-    provides_dist = JSONField(default=list)
-    obsoletes_dist = JSONField(default=list)
-    requires_external = JSONField(default=list)
-    classifiers = JSONField(default=list)
+    requires_dist = models.JSONField(default=list)
+    provides_dist = models.JSONField(default=list)
+    obsoletes_dist = models.JSONField(default=list)
+    requires_external = models.JSONField(default=list)
+    classifiers = models.JSONField(default=list)
+    project_urls = models.JSONField(default=dict)
+    description_content_type = models.TextField()
 
     def __str__(self):
         """
@@ -135,7 +179,7 @@ class PythonPackageContent(Content):
 
     class Meta:
         default_related_name = "%(app_label)s_%(model_name)s"
-        unique_together = ('filename',)
+        unique_together = ("sha256",)
 
 
 class PythonPublication(Publication):
@@ -159,9 +203,10 @@ class PythonRemote(Remote):
     """
 
     TYPE = 'python'
+    DEFAULT_DOWNLOAD_CONCURRENCY = 10
     prereleases = models.BooleanField(default=False)
-    includes = JSONField(default=list)
-    excludes = JSONField(default=list)
+    includes = models.JSONField(default=list)
+    excludes = models.JSONField(default=list)
     package_types = ArrayField(models.CharField(max_length=15, blank=True),
                                choices=PACKAGE_TYPES, default=list)
     keep_latest_packages = models.IntegerField(default=0)
@@ -181,5 +226,29 @@ class PythonRepository(Repository):
     CONTENT_TYPES = [PythonPackageContent]
     REMOTE_TYPES = [PythonRemote]
 
+    autopublish = models.BooleanField(default=False)
+
     class Meta:
         default_related_name = "%(app_label)s_%(model_name)s"
+
+    def on_new_version(self, version):
+        """
+        Called when new repository versions are created.
+
+        Args:
+            version: The new repository version
+        """
+        super().on_new_version(version)
+
+        # avoid circular import issues
+        from pulp_python.app import tasks
+
+        if self.autopublish:
+            tasks.publish(repository_version_pk=version.pk)
+
+    def finalize_new_version(self, new_version):
+        """
+        Remove duplicate packages that have the same filename.
+        """
+        remove_duplicates(new_version)
+        validate_repo_version(new_version)
